/*
 *  Switchbox v1.3 - controls 8 channels  6A. sigma I = 8A.
 *
 *  (c) 06/2006 gregor dorfbauer
 *  thanks to:
 *  peter dannegger - debounce library
 *  michael arndt - i2c-lcd-functions
 *  pascal stang - procyon avrlib
 */

#define UART_TX_BUFFER_SIZE 0x0001
#define UART_RX_BUFFER_SIZE 20

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

#include "switchbox_i2c.h"

#include "iocompat.h"
#include "lcd.h"
#include "i2c.h"
#include "timer.h"
#include "uart.h"


#define PROGRAM_LEN 20
#define CHANNEL_LEN 15
#define NAME_LEN	24

// Gertetasten 
#define KEY0		0
#define KEY1		1
#define KEY2		2
#define KEY3		3
#define KEY4		4
#define KEY5		5
#define KEY6		6
#define KEY7		7

//Funktionstasten unterhalb des LCD
#define FUNC0		8
#define FUNC1		9
#define FUNC2		10
#define FUNC3		11

//Funktionstasten haben repeat (Long-Press)

#define REPEAT_MASK		(1<<FUNC0^1<<FUNC1^1<<FUNC2^1<<FUNC3)	
#define REPEAT_START	25		// after 500ms
#define REPEAT_NEXT		10		// every 200ms

#define DISPLAY_DELAY 200UL		//~4sec question

#define SWITCH_OFF_BLINK 5		// - 100ms high, 100ms low. ~5 Hz 

#define SELFTEST_DELAY 100		//in ms

#define SWITCH_ON_DELAY 10		// ~1sec pause between channels

#define CLEAN 		10
#define TOBECLEANED	11
#define DIRTY 		12

#define VERSION_STRING "SWB1.3"

void getSwitchboxSettings();
void setSwitchboxSettings();

u08 EEPROM_DAS_ERSTE_BYTE_FUNKTIONIERT_NICHT EEMEM = 0;

u08 switchboxName[NAME_LEN+1] EEMEM = "     LAWGIVER  2006     ";

u08 szChannels[8][CHANNEL_LEN + 1] EEMEM  = 
{   "Test---Licht",
	"E-Herd", "Mikrowelle", "Toaster",
	"Entsafter", "Brotmaschine", "Eisfach", "Kaffeem."};

u08 szPrograms[4][PROGRAM_LEN + 1] EEMEM = 
{ "Alles ein", "Alles aus", "0x55", "0xAA" };

u08 eePrograms[4] EEMEM = { 0xFF, 0x00, 0x55, 0xAA };


u16 key_state;				// debounced and inverted key state:
							// bit = 1: key pressed
u16 key_press;				// key press detect

u16 key_rpt;				// key long press and repeat

volatile u08 screen;		// state of LCD (CLEAN, TOBECLEANED, DIRTY)

volatile u08 ssr_state;		//solid state relais
volatile u08 led_state;		//status leds

volatile u16 switch_off;	//bits to switch off (question)

volatile u08 current_program;	//filled from saved programs

volatile u08 switch_on_count;	//needed for staggered switch on / bitcounter

volatile u08 manual_override;	//overrides program settings (switch on sth forever...)

volatile u08 timeout;

void printStatus(void)
{
	lcdClear();


	lcdPrintEEStr(switchboxName);

	/*	
	lcd4BitOut(0,1);
	lcd4BitOut(1,1);
	lcd4BitOut(2,1);
	lcd4BitOut(3,1);
	lcd4BitOut(0,1);
	lcd4BitOut(0,1);	
*/
	lcd4BitOut(0x80 | 0x40, 0);	//next line
	if(current_program == 0)
		lcdPrintProgStr(PSTR("     select program"));

	screen = CLEAN;
}

/* knight-rider-testrun + one flash
 * called while booting.
*/
void led_SelfTest(void)
{	
	u08 c = 0;
	for(c=1;c<0x80;c<<=1)
	{
		c=~c;
		i2cMasterSend(I2C_STATUS_LED,1,&c);
		timerPause(SELFTEST_DELAY);
		c=~c;
	}
	c=~0x80;
	i2cMasterSend(I2C_STATUS_LED,1,&c);
	timerPause(SELFTEST_DELAY);

	for(c=0x40;c>0x00;c>>=1)
	{
		c=~c;
		i2cMasterSend(I2C_STATUS_LED,1,&c);
		timerPause(SELFTEST_DELAY);
		c=~c;
	}

	c=0x00;
	i2cMasterSend(I2C_STATUS_LED,1,&c);
	timerPause(SELFTEST_DELAY + 50);

	c=0xFF;
	i2cMasterSend(I2C_STATUS_LED,1,&c);
}

/* key debouncing. attached to timer0-overflow 
*/
void debounce(void)
{
	static u16 ct0, ct1, rpt;
	u16 i;
	u08 c;


	i2cMasterReceiveNI(I2C_STATUS_TAST,1,&c);
	i = c;
	i2cMasterReceiveNI(I2C_FUNCTION,1,&c);	
	i |= (c<<8);

	i = key_state ^ ~i;		// key changed ?
	ct0 = ~( ct0 & i );			// reset or count ct0
	ct1 = ct0 ^ (ct1 & i);		// reset or count ct1
	i &= ct0 & ct1;			// count until roll over ?
	key_state ^= i;			// then toggle debounced state
	key_press |= key_state & i;		// 0->1: key press detect

	if( (key_state & REPEAT_MASK) == 0 )	// check repeat function
		rpt = REPEAT_START;		// start delay
	if( --rpt == 0 ){
		rpt = REPEAT_NEXT;			// repeat delay
		key_rpt |= key_state & REPEAT_MASK;
	}

	if( (timer0GetOverflowCount() % SWITCH_OFF_BLINK) == 0 )	//blinken bei switch off
	{
		led_state ^= switch_off;
	}

	if( (switch_on_count < 8) &&
		(timer0GetOverflowCount() % SWITCH_ON_DELAY) == 0 )		//staggered switch on
	{

		while((switch_on_count < 8) &&
			((ssr_state & (1<<switch_on_count)) ||		//schon eingeschaltet
				!(current_program & (1<<switch_on_count)))//wird nicht eingeschaltet
			)	
		{
			switch_on_count++;
		}	
		ssr_state |= current_program & (1<<switch_on_count);
		led_state = ssr_state;
		switch_on_count++;
	}	

	if((timer0GetOverflowCount() > DISPLAY_DELAY))	//clean screen in main for.
	{
		timer0ClearOverflowCount();
		if(timeout != 0)
			timeout--;

		//	uartSendByte(timeout + '0');

		if(screen == DIRTY)
			screen = TOBECLEANED;
	}


}

u16 get_key_press( u16 key_mask )
{
	cli();					// read and clear atomic !
	key_mask &= key_press;                        // read key(s)
	key_press ^= key_mask;                        // clear key(s)
	sei();
	return key_mask;
}


u16 get_key_rpt( u16 key_mask )
{
	cli();					// read and clear atomic !
	key_mask &= key_rpt;                        	// read key(s)
	key_rpt ^= key_mask;                        	// clear key(s)
	sei();
	return key_mask;
}


u16 get_key_short( u16 key_mask )
{
	cli();			// read key state and key press atomic !
	return get_key_press( ~key_state & key_mask );
}


u16 get_key_long( u16 key_mask )
{
	return get_key_press( get_key_rpt( key_mask ));
}

void getSwitchboxSettings(void)
{
	u08 c,d;
	void *help = switchboxName;

	//switchbox title
	for(c=0;c<NAME_LEN;c++)
	{
		uartSendByte(eeprom_read_byte(help++));
	}
	//program titles
	for(d=0;d<4;d++)
	{
		help=szPrograms[d];
		for(c=0;c<PROGRAM_LEN;c++)
		{
			uartSendByte(eeprom_read_byte(help++));
		}
	}
	help=eePrograms;
	//program settings
	for(d=0;d<4;d++)
	{
		uartSendByte(eeprom_read_byte(help++));
	}
	for(d=0;d<8;d++)
	{
		help = szChannels[d];
		for(c=0;c<CHANNEL_LEN;c++)
		{
			uartSendByte(eeprom_read_byte(help++));			
		}
	}
}

void setSwitchboxSettings(void)
{
	u08 i = 0;
	u08 c = 0;
	u08 d = 0;

	void *help = switchboxName;

	//switchbox title
	for(i=0;i<NAME_LEN;i++)
	{
		while((timeout!=0) && !uartReceiveByte(&d));
		eeprom_write_byte(help++,d);
	}
	lcd4BitOut(0x80 | 0x40, 0);
	lcdPrintProgStr(PSTR("N"));
	//program titles
	for(i=0;i<4;i++)
	{
		help=szPrograms[i];
		for(c=0;c<PROGRAM_LEN;c++)
		{		
			while((timeout!=0) && !uartReceiveByte(&d));
			if(timeout == 0)
				goto to;

			eeprom_write_byte(help++,d);
		}
		lcdPrintProgStr(PSTR("P"));
	}
	help=eePrograms;
	//program settings
	for(i=0;i<4;i++)
	{
		while((timeout!=0) && !uartReceiveByte(&d));
		if(timeout == 0)
			goto to;

		eeprom_write_byte(help++,d);
		lcdPrintProgStr(PSTR("p"));
	}

	for(i=0;i<8;i++)
	{
		help = szChannels[i];
		for(c=0;c<CHANNEL_LEN;c++)
		{
			while((timeout!=0) && !uartReceiveByte(&d));
			if(timeout == 0)
				goto to;

			eeprom_write_byte(help++,d);
		}
		lcdPrintProgStr(PSTR("C"));
	}	
to:	
	if(timeout == 0)
	{
		lcdClear();
		lcdPrintProgStr(PSTR("timed out!"));

		timer0ClearOverflowCount();
		screen = DIRTY;
	}	
	else
	{
		lcdPrintProgStr(PSTR(" done."));
	}

}

int main (void)
{
	char version_string[] = VERSION_STRING;

	u08 c = 0;
	u08 d = 0;
	u08 i = 0;

	ssr_state = 0x0000;
	led_state = 0x00;

	switch_off = 0x00;
	current_program = 0x00;

	switch_on_count = 0xFF;

	manual_override = 0x00;

	timeout = 0;

	//let's read them from eeprom ...

	u08 programs[4] = { 0 };

	for(c=0;c<4;c++)
	{
		programs[c] = eeprom_read_byte(&eePrograms[c]);
	}


/*	
	u08 character[] = { 0x3,0x7,0xe,0xf,0x6,0xc,0x10,0x00 };
	u08 smiley[] = { 0x0,0xb,0x0,0x4,0x4,0x11,0xe,0x00 };
	u08 wolke[] = { 0x0,0x0,0xa,0x15,0x11,0x11,0xe, 0x00 };
	u08 bomb[] = {0x2,0x4,0xe,0x1f,0x1f,0x1f,0xe,0x00};
*/	
	//display_delay = 0;

	i2cInit();

	timerInit();

	timer0Init();

	timer0SetPrescaler(TIMER_CLK_DIV1024);	

// initialize the UART (serial port)
	uartInit();

	// set the baud rate of the UART for our debug/reporting output
	uartSetBaudRate(9600);


	for(c=0;c<sizeof(version_string);c++)
	{
		uartSendByte(version_string[c]);
	}	

	uartSendByte('\r');
	uartSendByte('\n');
	uartSendByte('4');
	//	uartSendByte('2');

/* init the lcd */
	lcdInit();
/*
	for(i=0;i<8;i++)
	{
		lcd4BitOut(0x40|i, 0);	//next line
		lcd4BitOut(character[i],1);
	}
	for(i=0;i<8;i++)
	{
		lcd4BitOut(0x40|(i+8), 0);	//next line
		lcd4BitOut(wolke[i],1);
	}
	for(i=0;i<8;i++)
	{
		lcd4BitOut(0x40|(i+16), 0);	//next line
		lcd4BitOut(smiley[i],1);
	}
	for(i=0;i<8;i++)
	{
		lcd4BitOut(0x40|(i+24), 0);	//next line
		lcd4BitOut(bomb[i],1);
	}
*/

	/* clear the display */
	//	lcdClear();

	//status output
	//	printStatus();

	timerAttach(TIMER0OVERFLOW_INT, debounce);	

	sei();



	//all outputs cleared (0xFF)
	c=0xFF;
	i2cMasterSendNI(I2C_SOLIDSTATE,1,&c);

	c=0xFF;		//off
	i2cMasterSendNI(I2C_STATUS_LED, 1, &c);

	//activate read on PCF8574
	c=0xFF;		//low-active
	i2cMasterSendNI(I2C_STATUS_TAST,1,&c);

	c=0xFF;
	i2cMasterSendNI(I2C_FUNCTION,1,&c);


	lcdClear();

	//	printStatus();
	lcdPrintProgStr(PSTR("Switchbox "));
	lcdPrintString(version_string);
	lcd4BitOut(0x80 | 0x40, 0);
	lcdPrintProgStr(PSTR(" (c) 2006 flo + gregor"));


	//fancy self test
	led_SelfTest();

	printStatus();
	// do we need this cli() here?
	cli();
	c=0xFF;
	i2cMasterSendNI(I2C_STATUS_LED, 1, &c);
	sei();

	ssr_state = 0x00;
	while(1)
	{

		for(i=0;i<=7;i++)	// pushbutton check (8)
		{

			if( get_key_press( 1<<i ) )
			{
				if(ssr_state & (1<<i))		// ist jetzt eingeschaltet
				{
					lcdClear();
					//lcdPrintString("switch off \"mixer\"?");
					lcdPrintProgStr(PSTR("off? "));
					lcdPrintEEStr(szChannels[i]);

					//a					lcdPrintEEStr(szChannels[i]);

					//					lcdPrintString(szChannels[i]);
					//					lcdPrintProgStr(PSTR("?"));

					lcd4BitOut(0x80 | 0x40, 0);
					lcdPrintProgStr(PSTR("[YES]               [NO]"));

					//switch off if answer is yes...
					//ssr_state ^= (1<<i);

					if(switch_off & (1<<i))	// doppelklick - blinkt schon
					{
						ssr_state ^= switch_off;
						current_program &= ~switch_off;
						manual_override &= ~switch_off;

						led_state = ssr_state;

						switch_off = 0;

						lcdClear();
						lcdPrintProgStr(PSTR("switched off..."));

					}
					else		// loss blinkn!
					{
						switch_off = (1<<i);
					}	
					// clear key buffer (function keys)
					// before asking the user
					get_key_press((1<<FUNC0) | (1<<FUNC1) | (1<<FUNC2) | (1<<FUNC3));

					timer0ClearOverflowCount();
					screen = DIRTY;
				}
				else					// ist jetzt ausgeschaltet
				{
					//lcdPrintString("0");
					lcdClear();
					//lcdPrintString("switch off \"mixer\"?");
					lcdPrintProgStr(PSTR("on "));
					//					lcdPrintString(szChannels[i]);
					lcdPrintEEStr(szChannels[i]);
					//					lcdPrintProgStr(PSTR("\""));

					timer0ClearOverflowCount();
					screen = DIRTY;

					manual_override |= (1<<i);
					ssr_state |= (1<<i);
					led_state |= (1<<i);

				}				
			}
		}

		if(switch_off != 0)		// should we switch something off?
		{
			if(get_key_press(1<<FUNC0))		// YES pressed
			{
				if(switch_off <= 0xFF)
				{
					ssr_state ^= switch_off;
					current_program &= ~switch_off;
					manual_override &= ~switch_off;
					led_state = ssr_state;
					switch_off = 0;

					lcdClear();
					lcdPrintProgStr(PSTR("switched off..."));

					timer0ClearOverflowCount();
					screen = DIRTY;
				}
				else
				{
					i = ((switch_off>>8)&0xF) - 1;

					while(switch_on_count < 8)
					{
						current_program &= ~(1<<switch_on_count);
						switch_on_count++;
					}

					if( (current_program == 0) ||
						(ssr_state != programs[i]) )
					{

						ssr_state ^= (current_program & (~programs[i]));
						ssr_state |= manual_override;

						led_state = ssr_state;
						switch_on_count = 0;		//start staggered switch on
						current_program = programs[i];
						timer0ClearOverflowCount();	//start power on immediately.
					}

					printStatus();
					lcd4BitOut(0x80 | 0x40, 0);	//next line
					//					lcdPrintString(szPrograms[i]);
					lcdPrintEEStr(szPrograms[i]);
					switch_off = 0;
				}

			}
			if(get_key_press(1<<FUNC3))		// NO pressed
			{
				switch_off = 0;
				led_state = ssr_state;

				lcdClear();
				lcdPrintProgStr(PSTR("done nothing..."));

				timer0ClearOverflowCount();
				screen = DIRTY;
			}
		}
		else	// handle function keys... no menu displayed.
		{
			for(i=0;i<4;i++)
			{
				if(get_key_short(1<<(FUNC0 + i)) || get_key_rpt(1<< ( FUNC0 + i)))
				{
					switch_off |= ((i+1)<<8);
					//					***********
					// change to selected program.
					lcdClear();
					//lcdPrintString("switch off \"mixer\"?");
					lcdPrintProgStr(PSTR("chg to: "));
					lcdPrintEEStr(szPrograms[i]);

					//a					lcdPrintEEStr(szChannels[i]);

					//					lcdPrintString(szChannels[i]);
					//					lcdPrintProgStr(PSTR(""));

					lcd4BitOut(0x80 | 0x40, 0);
					lcdPrintProgStr(PSTR("[YES]               [NO]"));

				}
			}

		}
		//clean screen
		if(screen == TOBECLEANED)
		{
			printStatus();
			switch_off = 0x00;
			led_state = ssr_state;
		}

		if(uartReceiveByte(&c))
		{
			switch(c)
			{
				case 's':
					lcdClear();
					lcdPrintProgStr(PSTR("receiving config..."));

					timer0ClearOverflowCount();
					screen = DIRTY;

					timeout=3;
					//					uartSendByte(timeout+'0');					

					setSwitchboxSettings();
				//get new programs
					for(c=0;c<4;c++)
					{
						programs[c] = eeprom_read_byte(&eePrograms[c]);
					}
					break;
				case 'g':
					lcdClear();
					lcdPrintProgStr(PSTR("sending config..."));

					timer0ClearOverflowCount();
					screen = DIRTY;

					getSwitchboxSettings();
					lcdPrintProgStr(PSTR("done."));
					break;
				case 'i':
					lcdClear();
					lcdPrintProgStr(PSTR("sending info..."));

					timer0ClearOverflowCount();
					screen = DIRTY;

					for(c=0;c<sizeof(version_string);c++)
					{
						uartSendByte(version_string[c]);
					}
					break;
				default:
					break;
			}
/*
			lcdClear();

			lcdPrintProgStr(PSTR("received something: "));
			lcd4BitOut(c,1);
			uartSendByte(c);
			*/
		}



		// set solid state relais status.

		//danke florian fr die logische verdrahtung!!!
		d=0;
		//		d = ssr_state;

		for(c=0;c<4;c++)
		{
			if(ssr_state & (1<<c))
				d |= (1<<(3-c));
		}

		for(c=4;c<8;c++)
		{
			if(ssr_state & (1<<c))
				d |= (1<<((7 - c) +4));
		}


		cli();
		c = ~d;
		i2cMasterSendNI(I2C_SOLIDSTATE,1,&c);	//solid state relais

		c = ~led_state;
		i2cMasterSendNI(I2C_STATUS_LED,1,&c);	//statusled
		sei();


	}
	return (0);
}
